/**
Functions to take exponents: b^n
Note: ^ is not c++ syntax for exponentiation
**/

#include <iostream>
#include <exception>
using namespace std;


// With a loop, accumulating up
double simpleExpLoopPart(int b, int n)
{
    double result = 1;

    for(int c = 1; c <= n; c++)
    {
        result *= b;
        cout << "c = " << c << ", and result = " << result << endl;
    }

    return result;
}; // what will happen if we enter a negative number? What about 0?


// With a loop and one-layer recursion for negatives
double simpleExpLoop(int b, int n)
{
    double result = 1;

    if(n >= 0)
    {
        for(int c = 1; c <= n; c++)
        {
            result *= b;
            cout << "c = " << c << ", and result = " << result << endl;
        }
    }
    else
    {
        return 1 / simpleExpLoop(b, -n);
    }

    return result;
};


// Fully recursively no negatives
double simpleExpRecurPart(int b, int n)
{
    cout << "Calling simpleExpRecurPart with b = "
        << b << " and n = " << n << endl;

    if(n == 0)
    {
        return 1;
    }
    else
    {
        return b * simpleExpRecurPart(b, n - 1);
    }

    // can have return statements above or 1 here
} // what will happen if we enter a negative number?


// Fully ecursively:
double simpleExpRecur(int b, int n)
{
    double result = 1;

    cout << "Calling simpleExpRecur with b = "
        << b << " and n = " << n << endl;

    if(n == 0)
    {
        return 1;
    }
    else if(n > 0)
    {
        return b * simpleExpRecur(b, n - 1);
    }
    else
    {
        return 1 / simpleExpRecur(b, -n);
    }

    // can have return statements above or 1 here
}


// Fast exponentiation (skipped 0^neg and negative special conditions here)
long long expEff(const long long b, const int n)
{
    cout << "Calling expEff with b = "
        << b << " and n = " << n << endl;
    if(n == 1) // this is a faster alternative to base case n==0 above, how much?
    {
        return b;
    }
    else if((n & 1) == 0) // bitwise faster than modulus to check even here
    {
        return expEff(b * b, n / 2);
    }
    else // odd
    {
        return expEff(b * b, n / 2) * b;
    }
    // can have return statements above or 1 here
}


int main()
{
    int b = 2;
    int n = 40;

    cout << "simpleExpLoopPart: " << simpleExpLoopPart(b, n) << endl << endl;
    cout << "simpleExpLoop: " << simpleExpLoop(b, n) << endl << endl;
    cout << "simpleExpRecurPart: " << simpleExpRecurPart(b, n) << endl << endl;
    cout << "simpleExpRecur: " << simpleExpRecur(b, n) << endl << endl;
    cout << "expEff: " << expEff(b, n) << endl << endl;

    return 0;
}

