/**
Various implementations of factorial
**/

#include <iostream>
#include "astack.h"
#include <exception>
using namespace std;


double factorial(int n, string indent = "")
{
    double result;

    cout << indent <<"Calling factorial with n = " << n << endl;

    if(n > 1)
    {
        // can just return instead if you don't want fancy print immediately below
        result = n * factorial(n - 1, indent += " ");

        cout << indent <<"Unwinding factorial with n = " << n
            << ",  and result = " << result << endl;
    }
    else if(n < 0) // undefined
    {
        throw exception();
    }
    else
    {
        result = 1;
    }
    return result;
}


double factorialTail(int n, int sum = 1, string indent = "")
{
    cout << indent <<"Calling factorial with n = "
        << n << " and sum = " << sum << endl;

    // error condition
    if(n < 0)
    {
        throw exception();
    }
    else if(0 == n || 1 == n)
    {
        return sum;
    }

    // Tail recursive call
    return factorialTail(n - 1, n * sum, indent += " ");

    // We can't cout anything on the way back up the stack
}


// Iterative Factorial using loops (imperative), accumulate pattern
double factorialLoop(int n)
{
    double total = 1;
    for(int i = 1; i < n+1; i++)
    {
        total *= i;
        cout << "FactLoop n = " << n
            << " and i = " << i
            << " and total = " << total << endl;
    }
    return total;
}

// three outside variables are referenced by the loop: total, i, and n

// Loop turned into a tail recursive function
double factLoopRec(int n, double total=1, int i=1)
{
    cout << "FactTail2 n = " << n
        << " and i = " << i
        << " and total = " << total << endl;
    if(i < (n + 1))
        return factLoopRec(n, total * i, i + 1);
    else
        return total;
}

// replacing recursion with a stack (uses stack class above)
// We'll cover this later
long long factStack(int n, Stack<double> &S)
{
    while(n > 1)
    {
        S.push(n--);
    }

    long long result = 1;

    while(S.length() > 0)
    {
        cout << "factStack n = " << n
            << " and result = " << result << endl;

        result = result * S.pop();
    }

    return result;
}


int main()
{
    int n;
    cout<<"Enter a number to find factorial: ";
    cin >> n;

    cout << "Factorial middle recursive: " << n
        << " = " << factorial(n) << endl << endl;

    cout << "Factorial tail recursive: " << n
        << " = " << factorialTail(n)<< endl << endl;

    cout << "Factorial loop: " << n
        << " = " << factorialLoop(n) << endl << endl;

    cout << "Factorial factLoopRec: " << n
        << " = " << factLoopRec(n=n) << endl << endl;

    AStack<double> stk(50);

    cout << "Factorial factStack: " << n
        << " = " << factStack(n, stk) << endl << endl;

    return 0;
}
