/**
Fibonacci
f(n) = f(n-2) + f(n-1)
     = 0 or 1 if n == 0 or n == 1
**/

#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
using namespace std;


// recursive Fibonacci (inefficient)
double fibonacci(int n)
{
    if(n == 0)
        return 0;
    else if(n == 1)
        return 1;
    else
        return fibonacci(n-1) + fibonacci(n-2);
}


double fibonacci2(int n)
{
    if(n < 2)
        return n;
    else
        return fibonacci(n-1) + fibonacci(n-2);
}


// fibonacci using a loop (efficient)
double fibonacciLoop(int n)
{
    int i;                    //the loop variable
    double previous = 0;      //the first Fibonacci number
    double current = 1;       //the second Fibonacci number

    for(i = 0; i < n; ++i)   //n steps
    {
        double temp = previous + current;
        previous = current;
        current = temp;
    }

    return previous;
}


// fib loop converted to tail recursive, is efficient
double fibConvert(int n, double previous = 0, double current = 1, int i = 0)
{
    if(i < n)
        return fibConvert(n, current, previous + current, i + 1);
    else
        return previous;
}


// Global cache...
int fibo_cache[100] = {0}; // this batch array initialization syntax only works for 0s

// memoization/caching improves efficiency
double fibonacciCached(int n)
{
    if(fibo_cache[n] == 0)
    {
        if(n < 2)
        {
            fibo_cache[n] = n;

            return n;
        }
        else
        {
            fibo_cache[n] = fibonacciCached(n-1) + fibonacciCached(n-2);
        }
    }

    return fibo_cache[n];
}


int main()
{
    for(int i = 1; i<44; i++)
        cout << "FibRec of " << i << " = " << fibonacci(i) << endl;

    for(int i = 1; i<44; i++)
        cout << "FibLoop of " << i << " = " << fibonacciLoop(i) << endl;

    for(int i = 1; i<44; i++)
        cout << "fibConvert of " << i << " = " << fibConvert(i) << endl;

    for(int i = 0; i<30; i++)
        cout << fibo_cache[i] << ", ";

    for(double i = 1; i<44; i++)
        cout << "fibonacciCached of " << i << " = " << fibonacciCached(i) << endl;

  return 0;
}
