/**
Tower of hanoi
**/

#include <iostream>
#include <string>
#include "astack.h"
using namespace std;


// do as I say, not as I do...
// these globals are just for easy display
AStack<string> sourceG(5);
AStack<string> targetG(5);
AStack<string> auxiliaryG(5);

void moveHanoi(int n,
        AStack<string> &source,
        AStack<string> &target,
        AStack<string> &auxiliary)
{
    if(n > 0)
    {
        // Move n-1 disks from source to auxiliary
        moveHanoi(n-1, source, auxiliary, target);

        // Move nth disk from source to target
        target.push(source.pop());

        // Display progress
        cout << "sourceG: " << sourceG << endl
            << "targetG: " << targetG << endl
            << "auxiliaryG: " << auxiliaryG << endl
            << endl;

        // Move n-1 disks left on auxiliary onto target
        moveHanoi(n-1, auxiliary, target, source);
    }
}


int hanoiCount(int n)
{
    if(n > 1)
    {
        return 2 * hanoiCount(n-1) + 1;
    }
    else return 1;
}


int main()
{

    /** What is the smallest number of steps to move the stack?
      In general: h_n = 2^n - 1, for all n >= 1,
      and this result is obtained by noting that steps 1 and 3 above take T_{n − 1} moves,
      and step 2 takes one move, giving T_h = 2T_{n − 1} + 1.*/
    for(int c = 1; c < 16; c++)
    {
        int countHanoi = hanoiCount(c);
        cout << "Moving" << c << " disks requires "
            << countHanoi << " moves." << endl;
    }

    // Define the starting point
    sourceG.push("Purple");
    sourceG.push("Blue");
    sourceG.push("Green");
    sourceG.push("Yellow");
    sourceG.push("Orange");

    cout << "sourceG: " << sourceG << endl
        << "targetG: " << targetG << endl
        << "auxiliaryG: " << auxiliaryG << endl
        << endl;

    // initiate call from source A to target C with auxiliary B
    moveHanoi(5, sourceG, targetG, auxiliaryG);

    return 0;
}
